#!/usr/bin/env python
# encoding: utf8

"""
A simple CLI for the automata package.
"""

import click
from .core import CellularAutomata


@click.group()
def cli():
    pass


@cli.command()
@click.option('--rule', type=int, required=True)
@click.option('--initial_conditions', type=str, required=True)
@click.option('--base', type=int, default=2)
@click.option('--frame_width', type=int, default=101)
@click.option('--frame_steps', type=int, default=200)
@click.option('--boundary_condition', type=click.Choice(['zero', 'periodic', 'one']), default='zero')
def generate(rule, initial_conditions, base, frame_width, frame_steps, boundary_condition):
    global automaton
    automaton = CellularAutomata(rule, initial_conditions, base, frame_width, frame_steps, boundary_condition)


# noinspection PyProtectedMember
@cli.command()
@click.option('--output_format', type=click.Choice(['png', 'jpg', 'svg']), default='png')
def display(output_format):
    if output_format == 'png':
        print(automaton._repr_png_())
    elif output_format == 'jpg':
        print(automaton._repr_jpeg_())
    elif output_format == 'svg':
        print(automaton._repr_svg_())


if __name__ == '__main__':
    cli()



# noinspection PyPackageRequirements
from IPython.display import display
# noinspection PyPackageRequirements
from ipywidgets import (IntSlider, Checkbox, ColorPicker, FloatSlider, Combobox, IntRangeSlider,
                        Layout, VBox, HBox, interactive)

from .core import Rule, SliceSpec, HighlightBounds, CellularAutomata

# USE - For use in Jupyter notebooks; assumes ipywidgets (and IPython) is installed in the Python environment


# TBD - Should be better coordinated with number of colors and color all_controls
_DISPLAY_BASE_MAX = 4

_MIN_SLICE_STEPS = 1


def get_controls(display_parameters=None, frame_steps=25, frame_width=201) -> dict:
    rule_slider = IntSlider(min=0, max=Rule(0, base=2).n_rules - 1, step=1, value=90, description='Rule:')
    base_slider = IntSlider(min=2, max=_DISPLAY_BASE_MAX, step=1, value=2, description='Base:')

    # Adjust the max r to correspond to the base
    def update_r_max(*_) -> None:
        rule_slider.max = Rule(0, base=int(base_slider.value)).n_rules - 1
    base_slider.observe(update_r_max, names='value')
    update_r_max()

    initial_conditions_entry = Combobox(
        value='1',
        options=[],
        description='Start:',
        ensure_option=False,  # allowing freeform input
        disabled=False
    )

    def validate_initial_conditions_entry(change):
        valid_chars = Rule.ALPHABET[:base_slider.value]
        new_value = ''.join([ch for ch in change['new'] if ch in valid_chars])

        # If the new value is different (i.e., some characters were filtered out),
        # set the Combobox value to '1'. Otherwise, add to options if not already present.
        if new_value != change['new']:
            initial_conditions_entry.value = new_value
        else:
            if new_value and new_value not in initial_conditions_entry.options:
                initial_conditions_entry.options = list(initial_conditions_entry.options) + [new_value]
    initial_conditions_entry.observe(validate_initial_conditions_entry, names='value')
    # validate_initial_conditions(chang=None)

    def validate_initial_conditions_for_base(*_):
        valid_chars = Rule.ALPHABET[:base_slider.value]
        if not all(ch in valid_chars for ch in initial_conditions_entry.value):
            initial_conditions_entry.value = '1'

    # Observe changes to the base_slider's value
    base_slider.observe(validate_initial_conditions_for_base, names='value')

    slice_slider = IntRangeSlider(
        value=[0, frame_steps],
        min=0,
        max=frame_steps,  # TBD - Change when frame_steps set by a control
        step=1,
        description='Slice:',
        disabled=False,
        continuous_update=True,
        orientation='horizontal',
        readout=True,
        readout_format='d',
    )

    def enforce_gap(change):
        old_lower, old_upper = change['old']
        lower, upper = change['new']

        if upper - lower < _MIN_SLICE_STEPS:
            # If the upper value has been moved most recently
            if upper != old_upper:
                new_lower = upper - _MIN_SLICE_STEPS
                # Ensure the new lower value isn't out of bounds
                if new_lower < slice_slider.min:
                    slice_slider.value = (slice_slider.min, slice_slider.min + _MIN_SLICE_STEPS)
                else:
                    slice_slider.value = (new_lower, upper)
            # If the lower value has been moved most recently
            elif lower != old_lower:
                new_upper = lower + _MIN_SLICE_STEPS
                # Ensure the new upper value isn't out of bounds
                if new_upper > slice_slider.max:
                    slice_slider.value = (slice_slider.max - _MIN_SLICE_STEPS, slice_slider.max)
                else:
                    slice_slider.value = (lower, new_upper)
    slice_slider.observe(enforce_gap, names=['value'])

    use_highlight_checkbox = Checkbox(value=False, description='Highlight')
    h_start_slider = IntSlider(min=0, max=frame_steps, step=1, value=0, description='Start:')
    h_width_slider = IntSlider(min=1, max=frame_width, step=1, value=21, description='Width:')
    h_offset_slider = IntSlider(min=-frame_width // 2, max=frame_width // 2, step=1, value=0,
                                description='Offset:')
    h_steps_slider = IntSlider(min=1, max=frame_steps, step=1, value=20, description='Steps:')

    rule_rows_slider = IntSlider(min=0, max=6, step=1, value=Rule(0, base=2).best_rows(),
                                 description='Rule rows:')

    def update_rule_rows_default(*_) -> None:
        rule_rows_slider.value = Rule(0, base=int(base_slider.value)).best_rows()

    base_slider.observe(update_rule_rows_default, names='value')
    update_rule_rows_default()

    # Enable/disable h_start and h_width based on the checkbox
    def update_highlight_controls(change=None) -> None:
        new_value = use_highlight_checkbox.value if change is None else change.new
        h_start_slider.disabled = not new_value
        h_width_slider.disabled = not new_value
        h_offset_slider.disabled = not new_value
        h_steps_slider.disabled = not new_value
    use_highlight_checkbox.observe(update_highlight_controls, names='value')
    update_highlight_controls()

    grid_color_picker = ColorPicker(concise=True, value='white', disabled=False, description='Grid color:')
    grid_thickness_slider = FloatSlider(min=0, max=2, step=0.025, value=0.2, description='Grid width:')

    cell_color_pickers = {0: ColorPicker(concise=True, value='black', disabled=False, description='0 ='),
                          1: ColorPicker(concise=True, value='yellow', disabled=False, description='1 ='),
                          2: ColorPicker(concise=True, value='red', disabled=False, description='2 ='),
                          3: ColorPicker(concise=True, value='green', disabled=False, description='3 =')}

    # Enable/disable cell color pickers based on current value of base
    def update_grid_color_controls(*_) -> None:
        for digit in cell_color_pickers.keys():
            if digit > int(base_slider.value) - 1:
                cell_color_pickers[digit].layout.display = 'none'
            else:
                cell_color_pickers[digit].layout.display = 'flex'  # REV - or 'block' or 'inline'

    base_slider.observe(update_grid_color_controls, names='value')
    update_grid_color_controls()

    controls = {
        'rule': rule_slider,
        'base': base_slider,
        'initial_conditions': initial_conditions_entry,
        'slice_bounds': slice_slider,
        'use_highlight': use_highlight_checkbox,
        'h_start': h_start_slider,
        'h_width': h_width_slider,
        'h_offset': h_offset_slider,
        'h_steps': h_steps_slider,
        'grid_color': grid_color_picker,
        'grid_thickness': grid_thickness_slider,
        'cell_color_0': cell_color_pickers[0],
        'cell_color_1': cell_color_pickers[1],
        'cell_color_2': cell_color_pickers[2],
        'cell_color_3': cell_color_pickers[3],  # REV - Can't be greater than _DISPLAY_BASE_MAX-1
        'rule_rows': rule_rows_slider
    }

    if display_parameters is None:
        display_parameters = controls.keys()

    # Filter all_controls based on the provided list of control names
    selected_controls = {parameter: controls[parameter] for parameter in display_parameters if parameter in controls}

    return selected_controls


def display_automaton(rule=90, base=2,
                      initial_conditions='1',
                      slice_bounds=None,
                      use_highlight=False, h_start=0, h_width=21, h_offset=0, h_steps=20,
                      grid_color='white', grid_thickness=0.2,
                      cell_color_0='black', cell_color_1='yellow', cell_color_2='red', cell_color_3='green',
                      rule_rows=1,
                      frame_steps=80, frame_width=151, fig_width=12) -> None:
    if not use_highlight:
        highlights = [HighlightBounds()]
    else:
        highlights = [HighlightBounds(steps=h_steps, start_step=h_start, offset=h_offset, width=h_width)]

    colors = [cell_color_0, cell_color_1, cell_color_2, cell_color_3]

    CellularAutomata(rule, initial_conditions, base=base,
                     frame_steps=frame_steps, frame_width=frame_width).display(
        CellularAutomata.DisplayParams(
            fig_width=fig_width,
            slice_spec=SliceSpec(slice_bounds[0], slice_bounds[1]-slice_bounds[0])
            if slice_bounds is not None else None,
            grid_color=grid_color, grid_thickness=grid_thickness,
            cell_colors=colors,
            highlights=highlights,
            check_highlight_bounds=False),
        rule_display_params=Rule.DisplayParams(cell_colors=colors, rows=rule_rows) if rule_rows > 0 else None,
        show_rule=rule_rows > 0
    )
    return


def interactive_display_automaton(frame_steps=80, frame_width=151, fig_width=12, display_parameters=None,
                                  controls=None) -> None:
    all_controls = get_controls(display_parameters=display_parameters,
                                frame_steps=frame_steps, frame_width=frame_width)

    # REV - Simpler version w/o layout
    # @interact(**all_controls)
    # def interactive_display(**kwargs):
    #     display_automaton(**kwargs, frame_steps=frame_steps, frame_width=frame_width, fig_width=fig_width)

    if controls is None:
        controls = all_controls.keys()

    def interactive_display(**kwargs):
        display_automaton(**kwargs, frame_steps=frame_steps, frame_width=frame_width, fig_width=fig_width)

    interactive_controls = interactive(interactive_display, **all_controls)

    def get_column_layout(control_number: int):
        return Layout(
            border='none',
            margin='0px',
            padding='0px',
            width='25%' if control_number > 0 else '0px',
            display='flex',
            flex_flow='column',
            align_items='flex-start',
            visibility='visible' if control_number > 0 else 'hidden'
        )

    output = interactive_controls.children[-1]

    all_column1_controls = ['rule', 'base', 'initial_conditions',
                            'cell_color_0', 'cell_color_1', 'cell_color_2', 'cell_color_3']
    all_column2_controls = ['slice_bounds', 'grid_color', 'grid_thickness', 'rule_rows']
    all_column3_controls = ['use_highlight', 'h_start', 'h_width', 'h_offset', 'h_steps']

    # column1_controls = filter(lambda item: item in controls, all_column1_controls)
    column1_controls = [c for c in controls if c in all_column1_controls]
    column2_controls = [c for c in controls if c in all_column2_controls]
    column3_controls = [c for c in controls if c in all_column3_controls]

    column1 = VBox([all_controls[name] for name in column1_controls],
                   layout=get_column_layout(control_number=len(column1_controls)))
    column2 = VBox([all_controls[name] for name in column2_controls],
                   layout=get_column_layout(control_number=len(column2_controls)))
    column3 = VBox([all_controls[name] for name in column3_controls],
                   layout=get_column_layout(control_number=len(column3_controls)))

    controls_layout = HBox([column1, column2, column3])

    display(controls_layout)
    display(output)



"""
A simple cellular automata based on those discussed in Wolfram's A New Kind of Science.
Currently limited to 1D, multi-state, immediate neighbor automata, using various boundary conditions.
"""

import io
import base64
from dataclasses import dataclass

import matplotlib.axes
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle
import matplotlib.gridspec as gridspec


class CellularAutomataError(ValueError):
    def __init__(self, *args: str):
        # super().__init__(", ".join(args).capitalize() + ".")
        super().__init__(", ".join(args) + ".")


@dataclass
class HighlightBounds:
    start_step: int = None
    steps: int = None
    offset: int = None
    width: int = None


@dataclass(frozen=True)
class SliceSpec:
    start_step: int = None
    steps: int = None

    def range(self):
        return slice(self.start_step, self.start_step + self.steps)


class Rule:
    def __init__(self, rule_number: int, base: int = 2):

        self.input_range = 1  # TBD - generalize to allow for larger neighborhoods, using argument
        # !!! - CellularAutomata currently makes no use of input_range; input_range of 1 is hard coded there

        self.base = base
        self.rule_number = rule_number

        self.input_span = 2 * self.input_range + 1        # Number of cells in the input neighborhood
        self.n_input_patterns = base ** self.input_span   # Number of input states for a rule
        self.n_rules = base ** self.n_input_patterns      # Number of possible rules for the given span and base

        # Validate the rule number
        if rule_number < 0 or rule_number > self.n_rules - 1:
            raise CellularAutomataError(f"Invalid rule number. Must be between 0 and {self.n_rules - 1}.")

        # Convert rule number to base representation; also the rule output pattern, in R-L order
        self.encoding = self._encode(self.rule_number, self.base, self.n_input_patterns)

        # Generate all configurations, of a given length
        input_patterns = [self._encode(i, self.base, length=self.input_span) for i in range(len(self.encoding))]
        input_patterns.reverse()  # Ordered to match rule number, which corresponds to output order by definition
        self.pattern_to_output = dict(zip(input_patterns, self.encoding))

        assert len(self.pattern_to_output) == self.n_input_patterns

    ALPHABET = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ'

    @staticmethod
    def get_cell_colors(color_list: [str], base: int) -> {str: str}:
        cell_colors = {e: c for e, c in
                       zip(Rule.ALPHABET[:base],
                           color_list if color_list else Rule.DisplayParams.get_default_cell_colors(base))}
        return cell_colors  # , cell_colors.values()

    @staticmethod
    def _encode(value: int, base: int, length: int) -> str:
        """
        Converts a value to its representation in the given base, padded to length.
        """
        if base > len(Rule.ALPHABET):
            raise ValueError(f"Base too large. Can't handle base > {len(Rule.ALPHABET)}")
        if base < 2:
            raise ValueError(f"Base too small. Can't handle base < 2")

        digits = [Rule.ALPHABET[value // base ** i % base] for i in range(length)][::-1]

        return '{:>{fill}{length}}'.format(''.join(digits), fill=Rule.ALPHABET[0], length=length)

    @staticmethod
    def _decode(encoding: str, base: int) -> int:
        """
        Converts a value in a given base to a base 10 value.
        """
        if base > len(Rule.ALPHABET):
            raise ValueError(f"Base too large. Can't handle base > {len(Rule.ALPHABET)}")
        if base < 2:
            raise ValueError(f"Base too small. Can't handle base < 2")
        if any(int(char) >= base for char in encoding if char.isdigit()):
            raise ValueError(f"Invalid value in encoding '{encoding}' for base {base}")

        return sum([Rule.ALPHABET.index(digit) * (base ** i) for i, digit in enumerate(reversed(encoding))])

    @dataclass(frozen=True)
    class DisplayParams:
        fig_width: float = 12
        gap: float = 0.2
        vertical_shift: float = 0.5
        cell_colors: [str] = None
        grid_color: str = 'black'
        grid_thickness: float = 0.5
        rows: int = 1

        @staticmethod
        def get_default_cell_colors(base):
            return [str(c) for c in np.linspace(1, 0, base)]

    def best_rows(self, max_per_row: int = 14) -> int:
        total = self.n_input_patterns
        return next(filter(lambda d: total % d == 0 and total // d <= max_per_row, range(1, total + 1)), total)

    def _plot_display(self, ax: matplotlib.axes.Axes, display_params: DisplayParams) -> None:

        cell_color_mapping = Rule.get_cell_colors(display_params.cell_colors, self.base)

        def draw_cell(x, y, d, cell_size=1):
            rect = Rectangle((x, y), cell_size, cell_size,
                             edgecolor=display_params.grid_color, facecolor=cell_color_mapping[d],
                             linewidth=display_params.grid_thickness)
            ax.add_patch(rect)

        # noinspection PyProtectedMember
        plot_params = _PlotParams(self.n_input_patterns, display_params.rows)
        ax.set_xlim(plot_params.x_min, plot_params.x_max)
        ax.set_ylim(plot_params.y_min, plot_params.y_max)
        ax.set_aspect('equal')
        ax.axis('off')

        for pattern_idx, input_pattern in enumerate(self.pattern_to_output.keys()):
            row_idx = pattern_idx // plot_params.patterns_per_row

            # Adjust y position based on row index
            y_top_adj = 2.5 * (display_params.rows - row_idx - 1) + 1.5
            y_bottom_adj = 2.5 * (display_params.rows - row_idx - 1) + 0.5

            # Adjust x position based on the center_offset
            x_shift = (pattern_idx % plot_params.patterns_per_row) * 4

            # Draw the input digits
            for input_digit_pos, input_digit in enumerate(input_pattern):
                draw_cell(input_digit_pos + x_shift, y_top_adj, input_digit)
            # Draw the output digit below the input, with a gap
            draw_cell(x_shift + 1, y_bottom_adj, self.pattern_to_output[input_pattern])

    def get_display(self, display_params: DisplayParams = None) -> (plt.Figure, (plt.Axes | None, plt.Axes | None)):
        return _get_display_grid(None, self,
                                 None, display_params,
                                 False, True)

    def display(self, display_params: DisplayParams = None) -> None:
        _, _ = self.get_display(display_params)
        plt.show()


class CellularAutomata:

    _instances = {}

    # Overriding the __new__ method for memoization
    def __new__(cls, *args, **kwargs):
        key = (args, frozenset(kwargs.items()))
        if key not in cls._instances:
            instance = super().__new__(cls)
            cls._instances[key] = instance
        return cls._instances[key]

    def __init__(self, rule_number: int, initial_conditions: str, base=2,
                 frame_width=101, frame_steps=200,
                 boundary_condition="zero"):

        # Avoid re-initializing an instance retrieved from memoization
        if getattr(self, '_initialized', False):
            return

        # Validate boundary condition
        valid_boundary_conditions = ["zero", "periodic", "one"]
        if boundary_condition not in valid_boundary_conditions:
            raise CellularAutomataError(
                f"Invalid boundary condition: {boundary_condition}. "
                f"Must be one of {', '.join(valid_boundary_conditions)}.")

        self.rule = Rule(rule_number, base)
        self.boundary_condition = boundary_condition

        self.frame_width = frame_width
        self.frame_steps = frame_steps

        self._lattice = np.empty((self.frame_steps, self.frame_width), dtype='<U1')
        self._lattice.fill(Rule.ALPHABET[0])

        # Process initial_conditions as a string and center on the 0th step
        # If the string length is even, pad it with a '0' at the left end
        if len(initial_conditions) % 2 == 0:
            initial_conditions = Rule.ALPHABET[0] + initial_conditions
        self.initial_conditions = initial_conditions

        center = self.frame_width // 2
        start = center - len(initial_conditions) // 2

        # Check for overflow of initial conditions
        if start < 0 or start + len(initial_conditions) > self.frame_width:
            raise CellularAutomataError("Initial conditions overflow the frame boundaries when centered.")

        for idx, char in enumerate(initial_conditions):
            if char not in Rule.ALPHABET[:base]:
                raise CellularAutomataError(f"Initial condition contains invalid symbol '{char}'.")
            self._lattice[0, start + idx] = char

        self._compute_automaton()

        self._initialized = True

    # Return the lattice, optionally with a slice specified using a SliceSpec object
    def lattice(self, slice_spec: SliceSpec = None) -> np.ndarray:
        slice_spec = self._validate_slice_bounds(slice_spec, check_bounds=True)
        return self._lattice[slice_spec.range()]

    def _get_boundary_values(self, current_row) -> (str, str):
        """
        Returns boundary values based on boundary condition and current row.
        """
        if self.boundary_condition == "zero":
            return Rule.ALPHABET[0], Rule.ALPHABET[0]
        elif self.boundary_condition == "periodic":
            return current_row[-1], current_row[0]
        elif self.boundary_condition == "one":
            return Rule.ALPHABET[1], Rule.ALPHABET[1]

    def _compute_automaton(self) -> None:
        """
        Generates the automaton based on the rule and initial conditions using a vectorized approach.
        """
        for row in range(1, self.frame_steps):
            # Get boundary values and extend current row
            left_boundary, right_boundary = self._get_boundary_values(self._lattice[row - 1])
            extended_current_row = np.hstack(([left_boundary], self._lattice[row - 1], [right_boundary]))

            # Use slicing to get left, center, and right neighbors for each cell
            left_neighbors = extended_current_row[:-2]
            center_neighbors = extended_current_row[1:-1]
            right_neighbors = extended_current_row[2:]

            # Form patterns considering the current base
            patterns = [left + center + right for left, center, right in
                        zip(left_neighbors, center_neighbors, right_neighbors)]

            # Use the patterns to directly get the output values from the rule's pattern_to_output dictionary
            self._lattice[row] = [self.rule.pattern_to_output[pattern] for pattern in patterns]

    def _validate_highlight_bounds(self, highlight: HighlightBounds, check_bounds: bool) -> None:
        """
        Checks bounds for highlighting and returns error messages for any explicitly provided invalid bounds.
        Set any unspecified highlight bounds to default values.
        Adjust unchecked provided bounds implement the intended highlight (which may be clipped or outside the frame).
        """
        # Check if the highlight region specified by any provided highlight bounds exceeds the bounds of the lattice
        if check_bounds:
            error_messages = []
            if all([highlight.start_step, highlight.steps]):
                if highlight.start_step + highlight.steps > self.frame_steps:
                    error_messages.append(
                        f"highlight starts at step {highlight.start_step} "
                        f"and ends at step {highlight.start_step + highlight.steps} (> {self.frame_steps})")
            if all([highlight.start_step]):
                if highlight.start_step < 0:
                    error_messages.append(f"highlight starts before step 0")
            if all([highlight.offset, highlight.width]):
                if (self.frame_width + highlight.width) // 2 + highlight.offset > self.frame_width:
                    error_messages.append(
                        f"highlight exceeds right bound with "
                        f"{(self.frame_width + highlight.width) // 2 + highlight.offset} (> {self.frame_width})")
                if (self.frame_width - highlight.width) // 2 + highlight.offset < 0:
                    error_messages.append(
                        f"highlight exceeds left bound with "
                        f"{(self.frame_width - highlight.width) // 2 + highlight.offset} (< 0)")
            if error_messages:
                raise CellularAutomataError(*error_messages)

        # Set any unspecified highlight bounds to default values
        if highlight.start_step is None:
            highlight.start_step = 0
        if highlight.steps is None:
            highlight.steps = self.frame_steps - highlight.start_step
        if highlight.offset is None:
            highlight.offset = 0
        if highlight.width is None:
            highlight.width = self.frame_width

        # Adjust unchecked provided bounds
        if highlight.start_step < 0:
            highlight.steps = max(0, highlight.steps + highlight.start_step)
            highlight.start_step = 0

    def _validate_slice_bounds(self, slice_spec: SliceSpec, check_bounds: bool) -> SliceSpec:
        """
        Check bounds for slice and error messages for any invalid bounds.
        Return a SliceSpec that is valid for the lattice.
        """
        if slice_spec is None:
            return SliceSpec(0, self.frame_steps)

        if check_bounds:
            if slice_spec.start_step is not None:
                if slice_spec.start_step < 0:   # slice_spec.start_step is None or
                    raise CellularAutomataError(f"Invalid slice start step {slice_spec.start_step}. Must be >= 0.")
                elif slice_spec.start_step >= self.frame_steps:
                    raise CellularAutomataError(
                        f"Invalid slice start step {slice_spec.start_step}. Must be < {self.frame_steps}.")
            if slice_spec.steps is not None:
                if slice_spec.steps < 1:  # slice_spec.steps is None or
                    raise CellularAutomataError(f"Invalid slice steps {slice_spec.steps}. Must be >= 1.")
                else:
                    fixed_start_step = slice_spec.start_step if slice_spec.start_step is not None else 0
                    if slice_spec.steps > self.frame_steps - fixed_start_step:
                        raise CellularAutomataError(
                            f"Invalid slice steps {slice_spec.steps}. "
                            f"Must be <= {self.frame_steps - fixed_start_step}.")

        validated_start_step = slice_spec.start_step
        validated_steps = slice_spec.steps

        if validated_start_step is None or validated_start_step < 0:
            validated_start_step = 0
        elif slice_spec.start_step >= self.frame_steps:
            validated_start_step = self.frame_steps - 1
        if validated_steps is None or validated_steps < 1:
            validated_steps = 1
        elif validated_steps > self.frame_steps - validated_start_step:
            validated_steps = self.frame_steps - validated_start_step

        return SliceSpec(validated_start_step, validated_steps)

    # TBD - This should be frozen and validating the bounds of the slice should be handled differently
    @dataclass
    class DisplayParams:
        fig_width: float = 12
        highlights: [HighlightBounds] = (HighlightBounds(),)
        slice_spec: SliceSpec = None
        highlight_mask: float = 0.3
        grid_color: str = None
        grid_thickness: float = 0.5
        cell_colors: [str] = None
        check_highlight_bounds: bool = True

    def _plot_display(self, ax: matplotlib.axes.Axes, display_params: DisplayParams) -> None:
        """
        Displays the cellular automaton lattice
        optionally highlighting a frame within it and
        optionally showing a grid around the cells.
        """

        cell_color_mapping = Rule.get_cell_colors(display_params.cell_colors, self.rule.base)

        # REV - Assumes _validate_highlight_bounds and _validate_slice_bounds by caller
        # for highlight in display_params.highlights:
        #     self._validate_highlight_bounds(highlight, check_bounds=display_params.check_highlight_bounds)
        # display_params.slice_spec = self._validate_slice_bounds(display_params.slice_spec, check_bounds=False)

        # Create a mask for highlighting, constraining the highlighted region to the bounds of the lattice
        mask = np.ones(self._lattice.shape, dtype=float) * display_params.highlight_mask
        for highlight in display_params.highlights:
            mask[highlight.start_step:highlight.start_step + highlight.steps,
                 max(0, (self.frame_width - highlight.width) // 2 + highlight.offset):
                 min(self.frame_width, (self.frame_width + highlight.width) // 2 + highlight.offset)] = 1

        # REV - See if there's a better way to do this
        # Convert the encoded lattice data to color using the color's dictionary.
        color_strings = np.vectorize(cell_color_mapping.get)(self._lattice[display_params.slice_spec.range()])
        # Convert the entire color_strings array to an array of RGBA values
        rgba_values = mpl.colors.to_rgba_array(color_strings.ravel())
        # Reshape the rgba_values to match the shape of the color_strings array with an additional dimension for RGBA
        rgb_lattice = rgba_values.reshape(*color_strings.shape, 4)
        # Mask the highlighted region
        rgb_lattice[..., 3] *= mask[display_params.slice_spec.range()]

        ax.imshow(rgb_lattice, aspect='equal', interpolation='none')

        # Add grid with lines around each cell if grid_color is specified
        if display_params.grid_color and display_params.grid_thickness > 0:
            ax.grid(which='minor', color=display_params.grid_color, linewidth=display_params.grid_thickness)
            ax.set_xticks(np.arange(-.5, self.frame_width, 1), minor=True)
            # Note: not using frame_steps here; only showing the slice
            ax.set_yticks(np.arange(-.5, display_params.slice_spec.steps, 1), minor=True)
            ax.grid(which='major', visible=False)
            ax.tick_params(which='both', bottom=False, left=False, labelbottom=False, labelleft=False)
        else:
            ax.axis('off')

        plt.setp(ax.spines.values(), color=display_params.grid_color, linewidth=display_params.grid_thickness)

    def get_display(self, display_params: DisplayParams = None,
                    rule_display_params: Rule.DisplayParams = None,
                    show_rule: bool = False) -> (plt.Figure, (plt.Axes | None, plt.Axes | None)):
        return _get_display_grid(self, self.rule,
                                 display_params, rule_display_params,
                                 True, show_rule)

    def display(self, display_params: DisplayParams = None,
                rule_display_params: Rule.DisplayParams = None,
                show_rule: bool = False) -> None:
        _, _ = self.get_display(display_params, rule_display_params, show_rule)
        plt.show()

    def to_dict(self):
        automaton_dict = {i: value for i, value in enumerate([''.join(row) for row in self._lattice])}
        automaton_dict['args'] = {
                "rule_number": self.rule.rule_number,
                "base": self.rule.base,
                "frame_width": self.frame_width,
                "frame_steps": self.frame_steps,
                "boundary_condition": self.boundary_condition,
                "initial_conditions": self.initial_conditions
            }
        return automaton_dict

    def _repr_png_(self):
        return _get_repr_img(self, None, True, False, 'png')

    def _repr_jpeg_(self):
        return _get_repr_img(self, None, True, False, 'jpg')

    def _repr_json_(self):
        return self.to_dict()

    # REV - This is a blurry mess
    def _repr_svg_(self):
        return _get_repr_img(self, None, True, False, 'svg')

    # REV - Just a block that says 'Image'
    def _repr_html_(self):
        return _get_repr_img(self, None, True, False, 'html')

    def _repr_text_(self):
        return '\n'.join([''.join(row) for row in self._lattice])

    def __repr__(self):
        return '\n'.join([''.join(row) for row in self._lattice])


@dataclass(frozen=True)
class _PlotParams:
    x_min: float
    x_max: float
    y_min: float
    y_max: float
    patterns_per_row: int
    height_ratio: float

    def __init__(self, total_patterns: int, rows: int, left_buffer: float = 0.25, right_buffer: float = 0.75):
        patterns_per_row = -(-total_patterns // rows)
        height_ratio = (4 * patterns_per_row - right_buffer + left_buffer) / (2.5 * rows)
        x_min = -left_buffer
        x_max = 4 * patterns_per_row - right_buffer
        y_min = 0
        y_max = 2.5 * rows

        object.__setattr__(self, 'patterns_per_row', patterns_per_row)
        object.__setattr__(self, 'height_ratio', height_ratio)
        object.__setattr__(self, 'x_min', x_min)
        object.__setattr__(self, 'x_max', x_max)
        object.__setattr__(self, 'y_min', y_min)
        object.__setattr__(self, 'y_max', y_max)


def _get_display_grid(automaton: CellularAutomata | None, rule: Rule | None,
                      display_params: CellularAutomata.DisplayParams | None,
                      rule_display_params: Rule.DisplayParams | None,
                      show_automaton: bool, show_rule: bool) -> (plt.Figure, (plt.Axes | None, plt.Axes | None)):

    assert not show_automaton or (show_automaton and automaton is not None)
    assert not show_rule or (show_rule and rule is not None)

    if display_params is None:
        display_params = CellularAutomata.DisplayParams()

    if rule_display_params is not None:
        show_rule = True

    if show_automaton:
        for highlight in display_params.highlights:
            # noinspection PyProtectedMember
            automaton._validate_highlight_bounds(highlight, check_bounds=display_params.check_highlight_bounds)
        # noinspection PyProtectedMember
        display_params.slice_spec = automaton._validate_slice_bounds(display_params.slice_spec, check_bounds=False)

    fig_width = display_params.fig_width if show_automaton else rule_display_params.fig_width

    if show_rule:
        if rule_display_params is None:
            rule_display_params = Rule.DisplayParams(cell_colors=display_params.cell_colors)
        # noinspection PyProtectedMember
        h_rule = (fig_width / _PlotParams(rule.n_input_patterns, rule_display_params.rows).height_ratio)
    else:
        h_rule = 0

    if show_automaton:
        h_lattice = fig_width * display_params.slice_spec.steps / automaton.frame_width
    else:
        h_lattice = 0

    fig = plt.figure(figsize=(fig_width, h_rule + h_lattice))
    gs = gridspec.GridSpec(2, 1,
                           height_ratios=[h_rule, h_lattice], hspace=0.0005 * fig_width)

    if show_rule:
        rule_ax = fig.add_subplot(gs[0])
        # noinspection PyProtectedMember
        rule._plot_display(rule_ax, rule_display_params)
    else:
        rule_ax = None

    if show_automaton:
        lattice_ax = fig.add_subplot(gs[1])
        # noinspection PyProtectedMember
        automaton._plot_display(lattice_ax, display_params)
    else:
        lattice_ax = None

    return fig, (rule_ax, lattice_ax)


def _get_repr_img(automaton: CellularAutomata | None, rule: Rule | None,
                  show_automaton: bool, show_rule: bool,
                  display_format: str) -> str | bytes:
    fig, _ = _get_display_grid(automaton, rule,
                               display_params=CellularAutomata.DisplayParams(),
                               rule_display_params=Rule.DisplayParams() if show_rule else None,
                               show_automaton=show_automaton, show_rule=show_rule)

    if display_format in ['svg']:
        buf = io.StringIO()
    elif display_format in ['png', 'jpg', 'html']:
        buf = io.BytesIO()
    # noinspection PyUnboundLocalVariable
    fig.savefig(buf, format=display_format if display_format != 'html' else 'svg')
    buf.seek(0)
    plt.close(fig)

    if display_format == 'html':
        data = base64.b64encode(buf.getvalue()).decode('utf-8')
        return f'<img src="data:image/png;base64,{data}" />'
    else:
        return buf.getvalue()



"""
A simple Python (3.10+) implementation aimed at simulating some cellular automata, primarily those focused on by Stephan Wolfram in his book "A New Kind of Science".

Functionality is currently limited to 1D, multi-state, immediate neighbor automata using various boundary conditions.
"""

# USE - Not needed if specified in pyproject.toml
# __author__ = 'Roy Levien' ;

# See - http://www.python.org/dev/peps/pep-0440/
# ^v(0|[1-9][0-9]*)(\.(0|[1-9][0-9]*))((a|b|rc)(0|[1-9][0-9]*))?(\.(dev|post)(0|[1-9][0-9]*))?$
# Note 0 is valid but no leading zeros
# regexr.com/7mmq7
# USE - (breaking).(feature).(fix)
# See - https://py-pkgs.org/07-releasing-versioning.html#version-numbering
__release__ = '0.1'  # N(.N)*
# USE - To indicate status of release, can be on any branch; sync with classifiers in pyproject.toml
__pre_release__ = 'a2'  # aN | bN | rcN
# USE - For all commits on develop branch, never on main branch, increment after each commit (that publishes)
__suffix__ = ''  # .devN |  (.postN)
__version__ = __release__ + __pre_release__ + __suffix__

from .core import *
